#ifndef __containers_h__
#define __containers_h__
#include <string.h>
#include <stddef.h>
#include <stdlib.h>
#include <stdbool.h>
// Forward declaration of the string collection type
typedef struct _StringCollection StringCollection;
struct _StringEnumerator;

// Type definition of the compare function
typedef int (*CompareFunction)(void *elem1,void *elem2);

// Definition of the functions associated with this type.
typedef struct {
	// Returns the number of elements stored
	int (*GetCount)(StringCollection &SC);

	// Is this collection read only?
	int (*IsReadOnly)(StringCollection &SC);

	// Sets this collection read-only or unsets the read-only flag
	int (*SetReadOnly)(StringCollection &SC,int flag);

	// Adds one element at the end. Given string is copied
	int (*Add)(StringCollection &SC,char *newval);

	// Adds a NULL terminated table of strings
	int (*AddRange)(StringCollection &SC,char **newvalues);

	// Clears all data and frees the memory
	int (*Clear)(StringCollection &SC);

	//Case sensitive search of a character string in the data
	bool (*Contains)(StringCollection &SC,char *str);

	// Copies all strings into a NULL terminated vector
	char **(*CopyTo)(StringCollection &SC);

	//Returns the index of the given string or -1 if not found
	int (*IndexOf)(StringCollection &SC,char *SearchedString);

	// Inserts a string at the position zero.
	int (*Insert)(StringCollection &SC,char *);

	// Inserts a string at the given position
	int (*InsertAt)(StringCollection &SC,int idx,char *newval);

	// Returns the string at the given position
	char *(*IndexAt)(StringCollection &SC,int idx);

	// Removes the given string if found
	int (*Remove)(StringCollection &SC,char *);

	//Removes the string at the indicated position
	int (*RemoveAt)(StringCollection &SC,int idx);

	// Frees the memory used by the collection
	int (*Finalize)(StringCollection &SC);

	// Returns the current capacity of the collection
	int (*GetCapacity)(StringCollection &SC);

	// Sets the capacity if there are no items in the collection
	bool (*SetCapacity)(StringCollection &SC,int newCapacity);

	// Calls the given function for all strings. "Arg" is a used supplied argument
	// that can be NULL that is passed to the function to call
	void (*Apply)(StringCollection &SC,int (*Applyfn)(char *,void * arg),void *arg);

	// Calls the given function for each string and saves all results
	// in an integer vector
	int *(*Map)(StringCollection &SC,int (*Applyfn)(char *));

	// Pushes a string, using the collection as a stack
	int (*Push)(StringCollection &SC,char *str);

	// Pops the last string off the collection
	char * (*Pop)(StringCollection &SC);

	// Replaces the character string at the given position with a new one
	char *(*ReplaceAt)(StringCollection &SC,int idx,char *newval);

	// Returns whether the collection makes case sensitive comparisons or not
	int (*IsCaseSensitive)(StringCollection &SC);

	// Sets case sensitivity by comparisons
	int (*SetCaseSensitive)(StringCollection &SC,int newval);

	// Compares two string collections
	bool (*Equal)(StringCollection &SC1,StringCollection &SC2);

	// Copy a string collection
	StringCollection (*Copy)(StringCollection &SC);
} StringCollectionInterface;

// Definition of the String Collection type
struct _StringCollection {
	StringCollectionInterface *lpVtbl; // The table of functions
        int count;                  /* in element size units */
        char **contents;               /* The contents of the collection */
        int capacity;                /* in element_size units */
	unsigned int flags;             // Read-only or other flags
};

struct _StringEnumerator;
typedef struct _StringEnumeratorFunctions {
	int (*Next)(struct _StringEnumerator &);
	int (*Previous)(struct _StringEnumerator &);
	char *(*GetCurrent)(struct _StringEnumerator &);
	int (*SetCurrent)(struct _StringEnumerator &,int);
} StringEnumeratorFunctions;

typedef struct _StringEnumerator {
	StringEnumeratorFunctions *lpVtbl;
	StringCollection *SC;
	int CurrentIndex;
} StringEnumerator;

// This are the only exported functions from this module
StringCollection newStringCollection(int startsize=0);
char *operator[](StringCollection *,int);
char *operator[]=(StringCollection *SC,int idx,char *newval);
/* ------------------------------------------------------------------------
 * LIST Interface                                                          *
 * This describes the single linked list data structure functions          *
 *-------------------------------------------------------------------------*/
typedef struct _list_element {
	struct _list_element *Next;
	char Data[];
} list_element;

typedef struct _List List;
typedef struct {
        // Returns the number of elements stored
        int (*GetCount)(List *AL);

        // Is this collection read only?
        int (*IsReadOnly)(List &AL);

        // Sets this collection read-only or unsets the read-only flag
        int (*SetReadOnly)(List &AL,int flag);

        // Adds one element at the end. Given element is copied
        int (*Add)(List &AL,void *newval);

        // Clears all data and frees the memory
        int (*Clear)(List &AL);

        //Case sensitive search of a character string in the data
        bool (*Contains)(List &AL,void *str);

        // Copies all strings into a NULL terminated vector
        List (*Copy)(List &AL);

        //Returns the index of the given string or -1 if not found
        int (*IndexOf)(List &AL,void *SearchedElement);

        // Inserts a value at the position zero.
        int (*Insert)(List &AL,void *);

        // Inserts a value at the given position
        int (*InsertAt)(List &AL,int idx,void *newval);

        // Returns the list at the given position
        void *(*IndexAt)(List *AL,int idx);

        // Removes the given string if found
        int (*Remove)(List &AL,void *);

        //Removes the string at the indicated position
        int (*RemoveAt)(List &AL,int idx);

        // Frees the memory used by the collection
        int (*Finalize)(List &AL);

        // Pushes a string, using the collection as a stack
        int (*Push)(List &AL,void *str);

        // Pops the last string off the list
        list_element * (*Pop)(List &AL);

        // Replaces the character string at the given position with a new one
        void *(*ReplaceAt)(List &AL,int idx,void *newval);

	// Sorts the list
	List (*Sort)(List &l,int (*compar)(const void *,const void *));

	// Reverses the list
	List (*Reverse)(List &l);

	// Gets an element range from the list
	List (*GetRange)(List &l,int start,int end);

	// Compares two list (Shallow comparison)
	bool (*Equal)(List &l1,List &l2);

	// Add a list at the end of another
	List (*Append)(List &l1,List &l2);

	// Comparison function
	int (*CompareFn)(void * elem1, void *elem2);

	// Set The comparison function
	CompareFunction (*SetCompareFunction)(List &l,CompareFunction fn);
} ListInterface;

struct _List {
	ListInterface *lpVtbl;
	size_t ElementSize;
	int count;                    /* in elements units */
	list_element *Last;          /* The last item */
       	list_element *First;        /* The contents of the list start here */
        unsigned Flags;
};
List newList(int element_size=0);
list_element * operator[]=(List &l,size_t position,void *data);
list_element * operator[](List &l,int position);
/*-------------------------------------------------------------------------
 * inline functions
 *------------------------------------------------------------------------*/
list_element *inline GetFirst(List *l)
{
	return l->First;
}
list_element * inline GetNext(list_element *l)
{
	return l->Next;
}

list_element * inline GetLast(List *l)
{
	return l->Last;
}

//****************************************************************************
//*                                                                          *
//*           The array list                                                 *
//*                                                                          *
//****************************************************************************

// Forward declaration of the string collection type
typedef struct _ArrayList ArrayList;

// Definition of the functions associated with this type.
typedef struct {
	// Returns the number of elements stored
	int (*GetCount)(ArrayList &AL);

	// Is this collection read only?
	int (*IsReadOnly)(ArrayList &AL);

	// Sets this collection read-only or unsets the read-only flag
	int (*SetReadOnly)(ArrayList &AL,int flag);

	// Adds one element at the end. Given element is copied
	int (*Add)(ArrayList &AL,void *newval);

	// Adds a NULL terminated table of strings
	int (*AddRange)(ArrayList &AL,void **newvalues);

	// Clears all data and frees the memory
	int (*Clear)(ArrayList &AL);

	//Case sensitive search of a character string in the data
	bool (*Contains)(ArrayList &AL,void *str);

	// Copies all strings into a NULL terminated vector
	void **(*CopyTo)(ArrayList &AL);

	//Returns the index of the given string or -1 if not found
	int (*IndexOf)(ArrayList &AL,void *SearchedString);

	// Inserts a string at the position zero.
	int (*Insert)(ArrayList &AL,void *);

	// Inserts a string at the given position
	int (*InsertAt)(ArrayList &AL,int idx,void *newval);

	// Returns the string at the given position
	void *(*IndexAt)(ArrayList &AL,int idx);

	// Removes the given string if found
	int (*Remove)(ArrayList &AL,void *);

	//Removes the string at the indicated position
	int (*RemoveAt)(ArrayList &AL,int idx);

	// Frees the memory used by the collection
	int (*Finalize)(ArrayList &AL);

	// Returns the current capacity of the collection
	int (*GetCapacity)(ArrayList &AL);

	// Sets the capacity if there are no items in the collection
	bool (*SetCapacity)(ArrayList &AL,int newCapacity);

	// Calls the given function for all strings. "Arg" is a used supplied argument
	// that can be NULL that is passed to the function to call
	void (*Apply)(ArrayList &AL,int (*Applyfn)(void *,void * arg),void *arg);

	// Pushes a string, using the collection as a stack
	int (*Push)(ArrayList &AL,void *str);

	// Pops the last string off the collection
	void * (*Pop)(ArrayList &AL);

	// Replaces the character string at the given position with a new one
	void *(*ReplaceAt)(ArrayList &AL,int idx,void *newval);

} ArrayListInterface;

// Definition of the String Collection type
struct _ArrayList {
	ArrayListInterface *lpVtbl; // The table of functions
        int count;                  /* number of elements in the array */
        void **contents;               /* The contents of the collection */
        int capacity;                /* allocated space in the contents vector */
	unsigned int flags;             // Read-only or other flags
	size_t ElementSize;		// Size of the elements stored in this array.
};

// This are the only exported functions from this module
ArrayList newArrayList(size_t elementsize,int startsize=0);
void *operator[](ArrayList AL,int idx);
void *operator[]=(ArrayList *AL,int idx,void *newval);

//******************************************************************************
//
//                          ErrorHandling
//
//******************************************************************************
#define CONTAINER_ERROR_BADARG  (0x80070057)
#define CONTAINER_ERROR_NOMEMORY        (0x8007000E)
#define CONTAINER_ERROR_INDEX   -3
#define CONTAINER_ERROR_READONLY        -4
#define CONTAINER_ERROR_FILEOPEN	-5
void ContainerRaiseError(char *fnname,unsigned errorcode);
#endif
