/*------------------------------------------------------------------------
 Module:        strcollection.c
 Author:        jacob
 Project:       Containers
 State:
 Creation Date:
 Description:   Implements the string collection container.
                Added fixes from Gerome. Oct. 20 2005
------------------------------------------------------------------------*/
#include "containers.h"
static int GetCount( StringCollection &SC);
static int IsReadOnly( StringCollection &SC);
static int SetReadOnly( StringCollection &SC,int newval);
static int Add( StringCollection &SC,char *newval);
static int AddRange( StringCollection &SC,char **newvalues);
static int Clear( StringCollection &SC);
static bool Contains( StringCollection &SC,char *str);
static char **CopyTo( StringCollection &SC);
static int IndexOf( StringCollection &SC,char *SearchedString);
static int Insert( StringCollection &SC,char *);
static int InsertAt( StringCollection &SC,int idx,char *newval);
static char *IndexAt( StringCollection &SC,int idx);
static int Remove( StringCollection &SC,char *);
static int RemoveAt( StringCollection &SC,int idx);
static int Finalize( StringCollection &SC);
static int GetCapacity( StringCollection &SC);
static bool SetCapacity( StringCollection &SC,int newCapacity);
static void Apply(StringCollection &SC,int(*Applyfn)(char *,void *),void *arg);
static int *Map(StringCollection &SC,int (*Applyfn)(char *));
static int Push(StringCollection &SC,char *str);
static char *Pop(StringCollection &SC);
static char *ReplaceAt(StringCollection &SC,int idx,char *newval);
static int IsCaseSensitive(StringCollection &SC);
static int SetCaseSensitive(StringCollection &SC,int newval);
static bool Equal(StringCollection &SC1,StringCollection &SC2);
static StringCollection Copy(StringCollection &SC);
static StringCollectionInterface lpVtableSC = {
   GetCount, IsReadOnly, SetReadOnly, Add, AddRange, Clear, Contains,
   CopyTo, IndexOf, Insert, InsertAt, IndexAt, Remove, RemoveAt,
   Finalize, GetCapacity, SetCapacity, Apply, Map, Push, Pop, ReplaceAt,
   IsCaseSensitive,SetCaseSensitive,Equal,Copy//ML,
};

#ifndef DEFAULT_START_SIZE
#define DEFAULT_START_SIZE 20
#endif
#ifndef NO_GC
#include <gc.h>
#define MALLOC GC_malloc
#define FREE(a)
#else
#include <stdlib.h>
#define MALLOC malloc
#define FREE free
#endif

static char * _stdcall DuplicateString(char *str,char *fnname)
{
   if (str == NULL)
      return NULL;
   char *result = MALLOC(strlen(str)+1);
   if (result == NULL) {
      ContainerRaiseError(fnname,CONTAINER_ERROR_NOMEMORY);
      return NULL;
   }
   strcpy(result,str);
   return result;
}

#define SC_READONLY   1
#define SC_IGNORECASE   2
#define CHUNKSIZE 20

StringCollection  newStringCollection(int startsize=0)
{
   StringCollection result;

   memset(&result,0,sizeof(StringCollection));
   result.lpVtbl = &lpVtableSC;
   if (startsize > 0) {
      result.contents = MALLOC(startsize*sizeof(char *));
      if (result.contents == NULL) {
         ContainerRaiseError(__func__,CONTAINER_ERROR_NOMEMORY);
      }
      else {
         memset(result.contents,0,sizeof(char *)*startsize);
         result.capacity = startsize;
      }
   }
   return result;
}

static int GetCount(struct _StringCollection &SC)
{
   return SC.count;
}
static int IsReadOnly(struct _StringCollection &SC)
{
   return (SC.flags & SC_READONLY) ? 1 : 0;
}
static int SetReadOnly(struct _StringCollection &SC,int newval)
{
   int oldval = SC.flags & SC_READONLY ? 1 : 0;
   if (newval)
      SC.flags |= SC_READONLY;
   else
      SC.flags &= ~SC_READONLY;
   return oldval;
}

static int IsCaseSensitive(StringCollection &SC)
{
   return (SC.flags & SC_IGNORECASE) ? 0 : 1;
}

static int SetCaseSensitive(StringCollection &SC,int newval)
{
   //ML int result = (SC.flags & SC_IGNORECASE) ? 1 : 0;
   int result = (SC.flags & SC_IGNORECASE) ? 0 : 1;
   //ML if (!newval)
   if (newval)
   SC.flags &= ~SC_IGNORECASE;
   else
   SC.flags |= SC_IGNORECASE;
   return result;
}

static int Resize(struct _StringCollection &SC)
{
   int newcapacity = SC.capacity + CHUNKSIZE;
   char **oldcontents = SC.contents;
   SC.contents = MALLOC(newcapacity*sizeof(char *));
   if (SC.contents == NULL) {
      SC.contents = oldcontents;
      return 0;
   }
   memset(SC.contents,0,sizeof(char *)*newcapacity);
   memcpy(SC.contents,oldcontents,SC.count*sizeof(char *));
   SC.capacity = newcapacity;
   FREE(oldcontents);
   return 1;
}

static int Add(StringCollection &SC,char *newval)
{
   if (SC.flags & SC_READONLY)
      return -1;
   if (SC.count >= SC.capacity) {
      if (!Resize(SC))
         return 0;
   }

   if (newval) {
      SC.contents[SC.count] = DuplicateString(newval,__func__);
      if (SC.contents[SC.count] == NULL) {
         return 0;
      }
   }
   else
      SC.contents[SC.count] = NULL;
   SC.count++;
   return SC.count;
}

static int AddRange(struct _StringCollection & SC,char **data)
{
   int i = 0,c;
   if (SC.flags & SC_READONLY)
      return 0;
   c = SC.count;
   while (data[i] != NULL) {
      int r = Add(SC,data[i]);
      if (r <= 0) {
         void *p;
         while (SC.count > c) {
            p = Pop(SC);
            FREE(p);
         }
         return r;
      }
      i++;
   }
   return SC.count;
}

static int Clear(struct _StringCollection & SC)
{
   int oldval = SC.count;
   if (SC.flags & SC_READONLY)
      return 0;
   for (int i=0; i<SC.count;i++) {
      FREE(SC.contents[i]);
      SC.contents[i] = NULL;
   }
   SC.count = 0;
   return oldval;
}

static bool Contains(struct _StringCollection & SC,char *str)
{
   int c;
   if (str == NULL)
      return false;
   c = *str;
   if ((SC.flags & SC_IGNORECASE) == 0) {
      for (int i=0; i<SC.count;i++) {
         if (c == SC.contents[i][0] && !strcmp(SC.contents[i],str))
            return true;
      }
   }
   else {
      for (int i = 0; i<SC.count;i++) {
         if (!stricmp(SC.contents[i],str))
            return true;
      }
   }
   return false;//ML true;
}

static char **CopyTo(struct _StringCollection &SC)
{
   char **result = MALLOC((1+SC.count)*sizeof(char *));
   int i;
   if (result == NULL)
      return NULL;
   for (i=0; i<SC.count;i++) {
      result[i] = DuplicateString(SC.contents[i],__func__);
   }
   result[i] = NULL;
   return result;
}

static int IndexOf(struct _StringCollection &SC,char *str)
{
   int i;
   int (*cmpfn)(const char *,const char *);

   if (SC.flags & SC_IGNORECASE) {
      cmpfn = stricmp;
   }
   else cmpfn = strcmp;

   for (i=0; i<SC.count;i++) {
      if (!cmpfn(SC.contents[i],str)) {
         return i;
      }
   }
   return -1;
}

char * __declspec(naked) operator[](StringCollection *SC, int idx)
{
}


static char *IndexAt(struct _StringCollection &SC,int idx)
{
   if (idx >=SC.count || idx < 0)
      return NULL;
   return SC.contents[idx];
}
char *operator[](StringCollection SC,int idx)
{
        return IndexAt(SC,idx);
}

static int InsertAt(struct _StringCollection &SC,int idx,char *newval)
{
   if (SC.flags & SC_READONLY)
      return 0;
   if (SC.count >= (SC.capacity-1)) {
      if (!Resize(SC))
         return 0;
   }
   if (idx < 0 || idx > SC.count)
      return -1;
   if (idx == 0) {
      if (SC.count > 0)
      memmove(SC.contents+1,SC.contents,SC.count*sizeof(char *));
      SC.contents[0] = newval;
   }
   else if (idx == SC.count) {
      SC.contents[idx] = newval;
   }
   else if (idx < SC.count) {
      memmove(SC.contents+idx+1,SC.contents+idx,(SC.count-idx+1)*sizeof(char *));
      SC.contents[idx] = newval;
   }
   SC.count++;
   return SC.count;
}

static int Insert(struct _StringCollection &SC,char *newval)
{
   if (SC.flags & SC_READONLY)
      return 0;
   return InsertAt(SC,0,newval);
}

static int RemoveAt(struct _StringCollection &SC,int idx)
{
   if (idx >= SC.count || idx < 0 || (SC.flags & SC_READONLY))
      return -1;
   if (SC.count == 0)
      return -2;
   FREE(SC.contents[idx]);
   if (idx < (SC.count-1)) {
      memmove(SC.contents+idx,SC.contents+idx+1,
         (SC.count-idx)*sizeof(char *));
   }
   SC.contents[SC.count-1]=NULL;
   SC.count--;
   return SC.count;
}

static int Remove(struct _StringCollection &SC,char *str)
{
   int idx = IndexOf(SC,str);
   if (idx < 0)
      return idx;
   return RemoveAt(SC,idx);
}

static int Push(struct _StringCollection &SC,char *str)
{
   char *r;
   if (SC.flags&SC_READONLY)
      return 0;
   if (SC.count >= SC.capacity-1) {
      if (!Resize(SC))
         return 0;
   }
   r = DuplicateString(str,__func__);
   if (r == NULL)
      return 0;
   SC.contents[SC.count++] = r;
   return SC.count;
}

static char * Pop(struct _StringCollection &SC)
{
   char *result;
   if ((SC.flags&SC_READONLY) || SC.count == 0)
      return NULL;
   SC.count--;
   result = SC.contents[SC.count];
   SC.contents[SC.count] = NULL;
   return result;
}

static int Finalize(struct _StringCollection &SC)
{
   int result = SC.count;
   for (int i=0; i<SC.count;i++) {
      FREE(SC.contents[i]);
   }
   FREE(SC.contents);
   FREE(&SC);
   return result;
}

static int GetCapacity(struct _StringCollection &SC)
{
   return SC.capacity;
}

static bool SetCapacity(struct _StringCollection &SC,int newCapacity)
{
   if (SC.count != 0)
      return 0;
   FREE(SC.contents);
   SC.contents = MALLOC(newCapacity*sizeof(char *));
   memset(SC.contents,0,sizeof(char *)*newCapacity);
   SC.capacity = newCapacity;
   return 1;
}

static void Apply(struct _StringCollection &SC,int (*Applyfn)(char *,void *),void *arg)
{
   for (int i=0; i<SC.count;i++) {
      Applyfn(SC.contents[i],arg);
   }
}

static int *Map(struct _StringCollection &SC,int (*Applyfn)(char *))
{
   int *result = MALLOC(SC.count*sizeof(int));

   if (result == NULL)
      return NULL;
        for (int i=0; i<SC.count;i++) {
                result[i] = Applyfn(SC.contents[i]);
        }
   return result;
}

char * __declspec(naked) operator[]=(StringCollection *SC,int idx,char *newval)
{
}

/* gedd123@free.fr (gerome) proposed calling DuplicateString. Good
  suggestion */
static char *ReplaceAt(StringCollection &SC,int idx,char *newval)
{
   if (SC.flags & SC_READONLY)
      return NULL;
   if (idx < 0 || idx > SC.count)
      return NULL;
   FREE(SC.contents[idx]);
   char *r = DuplicateString(newval,__func__);
   if (r == NULL)
	return 0;
   SC.contents[idx] = r;
   return newval;
}
char *operator[]=(StringCollection SC,int idx,char *newval)
{
        return ReplaceAt(SC,idx,newval);
}

static bool Equal(StringCollection &SC1,StringCollection &SC2)
{
   int i;
   if (SC1.count != SC2.count)
      return false;
   if (SC1.count == 0)
      return true;
   for (i=0; i<SC1.count;i++) {
      if (strcmp(SC1.contents[i],SC2.contents[i]))
         return false;
   }
   return true;
}

static StringCollection Copy(StringCollection &SC)
{
   int i;
   StringCollection result = newStringCollection(SC.count);

   for (i=0; i<SC.count;i++) {
      if (Add(result,SC.contents[i]) <= 0) {
         Finalize(result);
         break;
      }
   }
   return result;
}

#ifdef TEST
#include <stdio.h>
static void PrintStringCollection(StringCollection *SC)
{
   printf("Count %d, Capacity %d\n",SC->count,SC->capacity);
   for (int i=0; i<SC->count;i++) {
      printf("%s\n",SC[i]);
   }
   printf("\n");
}
int main(void)
{
   StringCollection *SC = newStringCollection();
   char *p;
   SC->Add("Martin");
   SC->Insert("Jakob");
   if (!SC->Contains("Martin")) {
      printf("error, string not found!\n");
   }
   printf("Count should be 2, is %d\n",SC->GetCount());
   PrintStringCollection(SC);
   SC->InsertAt(1,"Position 1");
   SC->InsertAt(2,"Position 2");
   PrintStringCollection(SC);
   SC->Remove("Jakob");
   PrintStringCollection(SC);
   SC->Push("pushed");
   PrintStringCollection(SC);
   SC->Pop();
   PrintStringCollection(SC);
   p = SC[1];
   printf("Item position 1:%s\n",p);
   PrintStringCollection(SC);
}
#endif
